/*
  Tutanchamun, an arcade game
  Copyright (C) 1998  David Kastrup
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
#include <unistd.h>

#include "tutan.h"
#include "tutdisp.h"

#include <sys/time.h>

timetab_t timtab[] =
  {{0,shotl},{0,shoml},
   {0,shotr},{0,shomr},
   {0,moveem},{0,timedo},{0,prod}};

#define TIMTBL (sizeof(timtab)/sizeof(timtab[0]))

unsigned long delay_accumulator = 0;
struct timeval lasttv;

void startclock()
{
  delay_accumulator = 0;
  gettimeofday(&lasttv, NULL);
}

void catchup()
{
  struct timeval curtv, del;
  long elapsed;

  if (delay_accumulator > 100) {
				/* Accumulated more than 50 ms */
    gettimeofday(&curtv, NULL);
    if (curtv.tv_sec > lasttv.tv_sec + 1 - (curtv.tv_usec >= lasttv.tv_usec)||
	curtv.tv_sec < lasttv.tv_sec) {
      /* More than 1 second lag or illegal time. */
      /* Assume just 200 ms */
      elapsed = 200000;
    } else {
      elapsed = (curtv.tv_sec - lasttv.tv_sec)*1000000
	+ curtv.tv_usec - lasttv.tv_usec;
    }
    if (elapsed + 10000 < delay_accumulator*500) {
      del.tv_sec = 0;
      del.tv_usec = delay_accumulator*500-elapsed;
      if ((curtv.tv_usec += del.tv_usec) >= 1000000) {
	do {
	  curtv.tv_sec++;
	} while ((curtv.tv_usec -= 1000000) >= 1000000);
      }
      delay_accumulator = 0;
      d_sleep(&del);
    } else
      if (elapsed > delay_accumulator*500 + 150000) {
				/* Our CPU is too slow.  Let not accumulate
				   more than 150ms of slack. */
	delay_accumulator = 150000/500;
    }
    lasttv = curtv;
  }
}

void paus(int n)
{
  struct timeval tv;
  tv.tv_sec = n;
  tv.tv_usec = 0;
  catchup();
  d_sleep(&tv);
  startclock();
}


static unsigned timcnt;    

int delay(unsigned n)
{
  int c;
  int i;
  int action = 0;

  delay_accumulator += n;
  
  if ((timcnt += n) < 167)
    return 0;
  c = timcnt / 167;
  timcnt %= 167;
  for (i=0; i<TIMTBL; ++i) {
    if (timtab[i].timebyte == DEAD_TIMER)
      continue;
    if ((timtab[i].timebyte += c) >= 0)
      action = 1;
  }
  return action;
}

jmp_buf gamebuf;

int tutmain()
{
  int i;

  timtab[MOVBYT].timebyte = -1;
  timtab[PROBYT].timebyte = playfield[player].stage.speed;
  if (playfield[player].stage.time[1] != ' '
      || playfield[player].stage.time[2] != '0') {
    timtab[TIMBYT].timebyte = -12;
    timtab[SHLBYT].timebyte = timtab[SHRBYT].timebyte = -1;
  } else
    timtab[TIMBYT].timebyte = timtab[SHLBYT].timebyte =
      timtab[SHRBYT].timebyte = DEAD_TIMER;
  timtab[SMLBYT].timebyte = timtab[SMRBYT].timebyte = DEAD_TIMER;
  timcnt = 0;
  if ((i = setjmp(gamebuf)) != 0) {
    bgstagedestroy();
    return i;
  }
  startclock();
  for (;;) {
  recheck:
    for (i=0; i<TIMTBL; ++i) {
      if (timtab[i].timebyte < 0)
	continue;
      (*timtab[i].timecode)(&timtab[i].timebyte);
    }
    for (;;) {
      catchup();
      if (!demflg) {
	askfir();
	asksma();
      }
      for (i=0; i<TIMTBL; ++i) {
	if (timtab[i].timebyte >= 0)
	  goto recheck;
      }
      delay(1);
    }
  }
}

void timedo (int *p)
{
  char * const tim = playfield[player].stage.time;
  *p -= 12;
  if (tim[2] != '0') {
    if (--tim[2] == '0'
	&& tim[1] == ' ') {
      *p = timtab[SHLBYT].timebyte = timtab[SHRBYT].timebyte = DEAD_TIMER;
    }
  } else {
    tim[2] = '9';
    if (tim[1] != '0') {
      if (--tim[1] == '0'
	  && tim[0] == ' ')
	tim[1] = ' ';
    } else {
      tim[1] = '9';
      if (--tim[0] == '0')
	tim[0] = ' ';
    }
  }
  d_wrtime(player, tim);
}
